"""Image transformation functions mimicking Scikit-Image's ones but based on
OpenCV (much faster)"""

import numpy as np
import cv2

# Correspondence between interpolation orders and corresponding commands in cv2

orders = {0: cv2.INTER_NEAREST,
          1: cv2.INTER_LINEAR,
          3: cv2.INTER_CUBIC,
          None: cv2.INTER_LINEAR}  # the last one is default

def rotate(image, angle, resize=False, center=None, order=None):
    """Rotation function similar to skimage.transform.rotate()

    Parameters
    ----------
    - image: image array (numpy array or equivalent)
    - angle: counter-clockwise rotation angle in degrees
    - resize: if True, expand image size to fit rotated image entirely
    - center: tuple of coords of rotation center (ignored if resize is True)
    - order: interpolation order, can be:
        - 0 (nearest neighbor)
        - 1 (bilinear)
        - 3 (bicubic)
        (if None, defaults to bilinear)

    Output
    ------
    Rotated image array
    """

    sy, sx, *_ = image.shape  # size of image in pixels (note inversion y/x)

    try:
        interpolation_method = orders[order]
    except KeyError:
        raise ValueError(f'{order} not a valid interpolation order.')

    # Calculate size of new frame if necessary ---------------------------------------------

    if resize:

        theta = np.deg2rad(angle)
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)

        # Projection legnths of rotated sides on original x and y
        lxx = abs(sx * cos_theta)
        lxy = abs(sx * sin_theta)
        lyx = abs(sy * sin_theta)
        lyy = abs(sy * cos_theta)

        sx_new = lxx + lyx
        sy_new = lxy + lyy

        output_size = int(sx_new), int(sy_new)

    else:

        output_size = sx, sy

    # Calculate rotation matrix --------------------------------------------------------

    if resize or center is None:
        center = (sx / 2, sy / 2)

    M = cv2.getRotationMatrix2D(center, angle, scale=1)  # no rescaling considered here

    if resize:  # Add translation to matrix so that the result fits into the new frame
        tranlsation = (sx_new - sx) / 2, (sy_new - sy) / 2
        M[:, 2] += tranlsation  # The last column of the matrix is the translation vector

    # Create rotated image ------------------------------------------------------------

    image_rotated = cv2.warpAffine(image, M, dsize=output_size, flags=interpolation_method)

    return image_rotated



"""Cropping image tools and related functions."""


import matplotlib.pyplot as plt
import matplotlib.patches as patches
from drapo import Cursor, rinput


# ======================= IMCROP and related functions =======================


def _cropzone_draw(ax, cropzone, c='r', linewidth=2):
    """Draw cropzone on axes."""
    x, y, w, h = cropzone
    rect = patches.Rectangle((x - 1 / 2, y - 1 / 2), w, h, linewidth=linewidth,
                             edgecolor=c, facecolor='none')
    ax.add_patch(rect)
    ax.figure.canvas.draw()
    return rect


def imcrop(*args, cmap=None, c='r', closefig=True, cursor=None,
           draggable=False, message='Crop Image', ax=None):
    """Interactive (or not)image cropping function using Numpy and Matplotlib.

    The *args allow to use the function in the two following ways:

    Main parameters (*args)
    -----------------------

    Depending on how the function is called, the cropping is interactive
    (manual selection on image) or imperative (crop zone (x, y, w, h) as input):

    *INTERACTIVE*

    `img_crop, cropzone = imcrop(img)`
    Input --> image (numpy array or equivalent)
    Output --> tuple (cropped image, crop rectangle (x, y, w, h))

    *IMPERATIVE*

    `img_crop = imcrop(img, cropzone)`
    Input --> image (numpy array or equivalent), crop zone (x, y, w, h)
    Output --> cropped image

    Other optional parameters
    -------------------------

    - cmap: colormap to display image in matplotlib imshow

    - c: color of lines / cursors in interactive mode

    - closefig: if True (default), close figure at end of interactive selection

    - cursor: appears to help selection by default but not in draggable mode
      (but can be forced in draggable mode by setting it to true, or can be
      completely suppressed by setting it to False). Default: None.

    - draggable: if True, use a draggable rectangle instead of clicks
      (only in interactive mode, see above)

    - message: message to show as title of the matplotlib window
      (only in interactive mode, see above)

    - ax: if not None, image shown in the ax matplotlib axes
      (only in interactive mode, see above)

    Note: when selecting, the pixels taken into account are those which have
    their centers closest to the click, not their edges closest to the click.
    For example, to crop to a single pixel, one needs to click two times in
    this pixel (possibly at the same location). For images with few pixels,
    this results in a visible offset between the dotted lines plotted after the
    clicks (running through the centers of the pixels clicked) and the final
    rectangle which runs along the edges of all pixels selected.

    Contrary to the Matlab imcrop function, the cropped rectangle is really of
    the width and height requested (w*h), not w+1 and h+1 as in Matlab.
    """
    img = args[0]  # load image
    sy, sx, *_ = img.shape  # size of image in pixels

    if len(args) == 2:
        interactive = False
        xmin, ymin, w, h = args[1]
    else:
        interactive = True

    if interactive:  # Interactive Drawing of Crop Rectangle -----------------

        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure

        if img.ndim == 2:  # grayscale image, use grayscale colormap
            cmap = 'gray' if cmap is None else cmap

        ax.imshow(img, cmap=cmap)
        ax.set_title(message)
        ax.set_xlabel('Click 2 pts to define crop (opposite corners of rectangle)')

        # Manage cursor visibility depending on mode -------------------------

        if cursor is None:
            cursor = False if draggable else True

        if cursor:
            Cursor()

        # --------------------------------------------------------------------

        if draggable:

            x_min, y_min, _w, _h = rinput(c=c)

            x_max = x_min + _w
            y_max = y_min + _h

        else:

            clicks = []

            for i in range(2):  # two clicks for two corners

                [(x_click, y_click)] = plt.ginput(1)
                clicks.append((x_click, y_click))

                # now, draw for visual clues ---------------------------------

                x_draw, y_draw = round(x_click), round(y_click)

                # draw lines corresponding to click (the -1/2 are used so that
                # the lines extend to the edges of the pixels)
                ax.plot([-1 / 2, sx - 1 / 2], [y_draw, y_draw], ':', color=c)
                ax.plot([x_draw, x_draw], [-1 / 2, sy - 1 / 2], ':', color=c)

                fig.canvas.draw()

            [(x1, y1), (x2, y2)] = clicks

            x_min, x_max = sorted((x1, x2))
            y_min, y_max = sorted((y1, y2))

        # Now, get pixels correspongind to clicks (center of a pixel is a
        # round number)
        xmin, xmax, ymin, ymax = [int(round(z)) for z in (x_min, x_max,
                                                          y_min, y_max)]

        # Calculate witdh and height in pixels
        w = xmax - xmin + 1
        h = ymax - ymin + 1

        cropzone = xmin, ymin, w, h

        _cropzone_draw(ax, cropzone, c)

        if closefig:
            plt.pause(0.1)
            plt.close(fig)

    # Now, in all cases, crop image to desired dimensions --------------------

    img_crop = img[ymin: ymin + h, xmin: xmin + w]

    if not interactive:
        return img_crop
    else:
        return img_crop, cropzone



"""Image analysis based on iso-gray level contours."""


import numpy as np


# ========================= Custom Contour Exception =========================

class ContourError(Exception):
    pass

# ======================== General contour functions =========================


def contour_coords(contour, source='scikit'):
    """Extract x, y tuple of contour positions from contour data.

    Scikit has reversed x, y coordinates
    OpenCV has an unusual numpy array shape (npts, 1, 2)

    Parameters
    ----------
    contour: contour data
    source: 'scikit' or 'opencv'

    Output
    ------
    x, y (tuple of numpy arrays) that can be used directly on an imshow() graph
    """
    if source == 'scikit':
        return contour[:, 1], contour[:, 0]
    elif source == 'opencv':
        contour = contour.squeeze()   # eliminate middle dimension in array
        return contour[:, 0], contour[:, 1]
    else:
        raise ValueError(f'{source} not a valid source for contour data.')


def contour_properties(x, y):
    """Returns centroid position, perimeter and area of a contour (x, y, p, a)

    Parameters
    ----------
    x, y : coordinates of contour (arrays that must support math operations)

    Output
    ------
    Dictionary with keys:
    - 'centroid': typle (x, y) of centroid position
    - 'perimeter': contour length (float)
    - 'area': signed area (float)

    Example
    -------
    Test can be done with the following hexagon :

    l = 1 / np.sqrt(3)
    xp = np.array([1, 1, 0, -1, -1, 0])/2
    yp = np.array([-l, l, 2*l, l, -l, -2*l])/2

    x, y, p, a = contour_properties(xp, yp)
    should return
    x = 0, y = 0, p = 6/sqrt(3) ~ 3.4641, a = -sqrt(3)/2 ~ -0.8660

    (area is negative because the contour is anti-clockwise with respect to an
    imshow plot, and clockwise with respect to a standard plot.)
    """
    # shift data around approx. center of contour (useful?)
    xm, ym = np.mean(x), np.mean(y)
    x = x - xm
    y = y - ym

    # derivatives of x and y
    x_looped = np.append(x, x[0])   # loop in case contour is not closed
    y_looped = np.append(y, y[0])
    dx = np.diff(x_looped)
    dy = np.diff(y_looped)

    # signed area and perimeter are straightforward to calculate
    area = np.sum(y * dx - x * dy) / 2          # area
    perimeter = np.sum(np.sqrt(dx**2 + dy**2))  # perimeter

    # moments are needed for centroid position calculation
    Ma = np.sum((y * dx**2 - x**2 * dy) / 4 + x * y * dx / 2 + dx**2 * dy / 12)
    Mb = np.sum((y**2 * dx - x * dy**2) / 4 - x * y * dy / 2 - dy**2 * dx / 12)

    # centroid position
    xc = Ma / area + xm
    yc = Mb / area + ym

    return {'centroid': (xc, yc),
            'perimeter': perimeter,
            'area': area}


def closest_contour(contours, position, edge=False, source='scikit'):
    """Finds the closest contour (skimage) to a certain position (tuple x, y)

    Parameters
    ----------
    - contours: list of contours obtained from skimage or opencv analysis
    (must be numpy arrays)
    - position: tuple (x, y)
    - edge:
        * if True, returns the contour with the edge closest to the position
        * if False (default), returns the contour with the average position
        closest to position.
    - source: 'scikit' or 'opencv'

    Output
    ------
    Contour within the initial list that is the closest to position
    """
    if len(contours) == 0:
        raise ContourError('No Contours Available')

    if len(contours) == 1:  # just one contour, no need to sort contours
        contour, = contours
        return contour

    else:
        x, y = position
        dists = []  # will store distances to position

        for contour in contours:

            xcont, ycont = contour_coords(contour, source)

            # center mode: compare approx. center of contour to position
            if not edge:
                dists.append(np.hypot(x - xcont.mean(), y - ycont.mean()))
            # edge mode: compare closest point on contour edge to position
            else:
                localdists = np.hypot(xcont - x, ycont - y)
                dists.append(localdists.min())

        imin = dists.index(min(dists))  # returns only first index if several ok

        return contours[imin]



"""Tests for the imgbasics module (pytest), basic functions (imcrop, etc.)"""

# Standard library
from pathlib import Path

# Non standard modules
from skimage import io

# Local module
import imgbasics
from imgbasics.transform import rotate


# =============================== Misc. config ===============================


datafolder = Path(imgbasics.__file__).parent / '..' / 'data'

img = io.imread(datafolder / 'example.png')
img_color = io.imread(datafolder / 'example_color.png')


# ========================== Test transform module ===========================


def test_transform_rotate():
    """23 degrees rotation with frame resize to fit all rotated image."""
    img_rot = rotate(img, angle=-23, resize=True, order=3)
    assert img_rot.shape == (569, 608)


def test_transform_rotate_color():
    """Same but with a color image, and no resize (same size before/after)."""
    img_rot = rotate(img_color, angle=-23)
    assert img_rot.shape == img_color.shape



"""Tests for the imgbasics module (pytest), basic functions (imcrop, etc.)"""

# Standard library
from pathlib import Path

# Non standard modules
import numpy as np
import matplotlib.pyplot as plt
from skimage import io, measure

# Local module
import imgbasics
from imgbasics import contour_properties, closest_contour, imcrop, contour_coords
from imgbasics.cropping import _cropzone_draw


# =============================== Misc. config ===============================

datafolder = Path(imgbasics.__file__).parent / '..' / 'data'

# =========== Test contour properties calculations with a hexagon ============


def test_contprops():

    # Make hexagon
    lh = 1 / np.sqrt(3)
    xp = np.array([1, 1, 0, -1, -1, 0]) / 2
    yp = np.array([-lh, lh, 2 * lh, lh, -lh, -2 * lh]) / 2

    # Calculate properties of hexagon
    contprops = contour_properties(xp, yp)

    x, y = contprops['centroid']
    p = contprops['perimeter']
    a = contprops['area']

    assert round(x, 6) == 0
    assert round(y, 6) == 0
    assert round(p, 4) == 3.4641
    assert round(a, 4) == -0.8660


# =========== Test find closest contour to position on real image ============
# (also re-tests contour_properties on more realistic example)

img = io.imread(datafolder / 'example.png')
contours = measure.find_contours(img, 170)

def test_contours_1():

    c = closest_contour(contours, (221, 281))  # should select the lowest, white spot
    xs, ys = contour_coords(c, source='scikit')

    contprops = contour_properties(xs, ys)

    x, y = contprops['centroid']
    p = contprops['perimeter']
    a = contprops['area']

    assert round(x) == 232
    assert round(y) == 280
    assert round(p) == 54
    assert round(a) == -219


# ============== Test imcrop with a random numpy array as image ==============

x = 5
y = 7
w = 14
h = 9

img2 = np.random.randint(0, 256, (20, 20))

cropzone = x, y, w, h
img_crop = imcrop(img2, cropzone)


def test_imcrop_1():
    """Check that output of imcrop in non-interactive mode has correct shape."""
    assert img_crop.shape == (h, w)


def test_imcrop_2():
    """Check that interactive option returns the same."""
    fig, ax = plt.subplots()
    ax.imshow(img2)
    _cropzone_draw(ax, cropzone, c='b')
    msg = 'Redefine blue crop zone (click slightly inside of rectangle)'
    img_crop_2, cropzone_2 = imcrop(img2, ax=ax, message=msg)
    assert (img_crop_2 == img_crop).all()


def test_imcrop_3():
    """Check that interactive option with draggable rectangle does the same."""
    fig, ax = plt.subplots()
    ax.imshow(img2)
    _cropzone_draw(ax, cropzone, c='b')
    msg = 'Redefine blue crop zone by dragging rectangle'
    img_crop_2, cropzone_2 = imcrop(img2, draggable=True, ax=ax, message=msg)
    assert img_crop_2.shape == img_crop.shape


# ====================== Test imcrop with a color image ======================


def test_imcrop_color():
    """Check that cropping works ok with a color image as well."""
    w = 236
    h = 71
    zone = 34, 125, w, h
    imgc = io.imread(datafolder / 'example_color.png')
    imgcrop = imcrop(imgc, zone)
    hcrop, wcrop, *_ = imgcrop.shape
    assert (hcrop, wcrop) == (h, w)
